#include <stdlib.h>
#include <string.h>

#include "spline.h"

// Wertet den kubischen Spline s an der Stelle t \in [0,1] aus
p_t spline_eval (spline_t * s, double t)
{
    return (p_t) 
        {
            gsl_spline_eval (s->spline_x, t, s->acc_x), 
            gsl_spline_eval (s->spline_y, t, s->acc_y) 
        };
}

// Allokiert Speicher fr einen kubischen Spline mit n Punken
// Die Anschlussbedingung ist periodisch.
// Intern brauchen wir einen Punkt mehr damit es vom Ende
// zum Anfang keinen Sprung gibt und die Perioden stetig
// aufeinander folgen.
spline_t * spline_alloc (unsigned int n, int interpol)
{
    spline_t * s = (spline_t*) calloc (1, sizeof (spline_t));

    s->n = n;
    s->x = (double*) calloc (n+1, sizeof (double));
    s->y = (double*) calloc (n+1, sizeof (double));
    
    s->acc_x   = gsl_interp_accel_alloc ();
    s->acc_y   = gsl_interp_accel_alloc ();
    
    switch (interpol)
    {
        case INTERPOL_SPLINE:
            s->spline_x = gsl_spline_alloc (gsl_interp_cspline_periodic, n+1);
            s->spline_y = gsl_spline_alloc (gsl_interp_cspline_periodic, n+1);
            break;
            
        case INTERPOL_LINEAR:
            s->spline_x = gsl_spline_alloc (gsl_interp_linear, n+1);
            s->spline_y = gsl_spline_alloc (gsl_interp_linear, n+1);
            break;
    }
    
    return s;
}

// Den kubischen Spline s initialisieren.
// x und y sind die Koordinaten, die s interpolieren soll.
// x und y mssen Lnge n haben (mit dem der Spline allokiert wurde)
// l stellt eine Funktion 0..n -> [0,1] dar und liefert die t-Werte, 
// welche die Sttzstellen fr den Spline bilden, d.h. die x-Koordinate
// des Splines ist gegeben durch sie Sttzpunkte (l(k), x(k))
void spline_init (spline_t * s, lambda_t * l, double * x, double * y)
{
    unsigned int n = s->n;
    
    memcpy (s->x, x, n*sizeof (double));
    memcpy (s->y, y, n*sizeof (double));
    
    s->x[n] = s->x[0];
    s->y[n] = s->y[0];
    
    gsl_spline_init (s->spline_x, l->t, s->x, n+1);
    gsl_spline_init (s->spline_y, l->t, s->y, n+1);
}

void spline_free (spline_t * s)
{
    gsl_spline_free (s->spline_y);
    gsl_spline_free (s->spline_x);
    gsl_interp_accel_free (s->acc_y);
    gsl_interp_accel_free (s->acc_x);
    free (s->y);
    free (s->x);
    free (s);
}

// Zeichnet den Spline mit np Punkten
void spline_plot (spline_t * s, int color, unsigned int np)
{
    for (unsigned int k=0; k < np; k++)
    {
        double t = (double) k / np;
        dot (spline_eval (s, t), color, 1);
    }
}



