#include <stdlib.h>
#include <math.h>

#include <gsl/gsl_errno.h>
#include <gsl/gsl_fft_real.h>
#include <gsl/gsl_fft_halfcomplex.h>

#include <gsl/gsl_wavelet.h>

#include "error.h"
#include "trafo.h"
#include "lambda.h"
#include "ppm.h"

gsl_fft_real_wavetable * real;
gsl_fft_real_workspace * work;
static gsl_wavelet * wave;
static gsl_wavelet_workspace * wwork;
     
void fft_alloc (size_t n)
{
    real = gsl_fft_real_wavetable_alloc (n);
    work = gsl_fft_real_workspace_alloc (n);
     
    if (n & (n-1))
        error ("Wavelets brauchen 2-Poten, %d ist aber keine", n);
        
    wave  = gsl_wavelet_alloc (gsl_wavelet_daubechies, 20);
    //wave  = gsl_wavelet_alloc (gsl_wavelet_haar, 2);
    wwork = gsl_wavelet_workspace_alloc (n);
}

void fft_free (void)
{
    gsl_fft_real_wavetable_free (real);
    gsl_fft_real_workspace_free (work);
    gsl_wavelet_free (wave);
    gsl_wavelet_workspace_free (wwork);
}

trafo_t * trafo_alloc (unsigned int n)
{
    trafo_t * f = (trafo_t*) calloc (1, sizeof (trafo_t));
    
    f->n = n;
    
    f->x = (double*) calloc (n, sizeof (double));
    f->y = (double*) calloc (n, sizeof (double));
    
    f->cx = (double*) calloc (2*n, sizeof (double));
    f->cy = (double*) calloc (2*n, sizeof (double));
    
    return f;
}

p_t to_cos_sin (double * c, unsigned int k, unsigned int n)
{
    // c_k * exp (2 * Pi * i * k/N * t)
    p_t ck = { REAL(c,k), -IMAG(c,k) };
    
    if (k != 0 && 2*k != n)
    {
        ck.x += REAL(c,n-k);
        ck.y += IMAG(c,n-k);
    }
    
    if (k == 0)
        ck.y = 0.0;
        
    ck.x /= n;    
    ck.y /= n;    
        
    return ck;
}

// http://de.wikipedia.org/wiki/Additionstheorem#Sinusoid
p_t to_a_phi (double * c, unsigned int k, unsigned int n)
{
    // c_k * exp (2 * Pi * i * k/N * t)
    p_t ck = to_cos_sin (c, k, n);

    double r   = hypot (ck.x, ck.y);
    double phi = (r > 1e-5) ? atan2 (-ck.y, ck.x) : 0.0;
    
    return (p_t) { r, phi };
}

// FFT-Trafo an der Stelle t ausweten
// f ist eine Funktion R -> R^2 mit Periode N
p_t trafo_eval (trafo_t * f, double t)
{
    double x = 0.0;
    double y = 0.0;
    
    for (unsigned int k=0; k <= f->n - k; k++)
    {
        p_t c;
        double a = 2.0 * M_PI * k/f->n * t;
        
        c = to_cos_sin (f->cx, k, f->n);  x += c.x * cos(a) + c.y * sin(a);
        c = to_cos_sin (f->cy, k, f->n);  y += c.x * cos(a) + c.y * sin(a);
    }
    
    return (p_t) { x, y };
}

p_t trafo_eval_aphi (trafo_t * f, double t, double phi)
{
    double x = 0.0;
    double y = 0.0;
    
    for (unsigned int k=0; k <= f->n - k; k++)
    {
        p_t c;
        double a = 2.0 * M_PI * k/f->n * t;
        
        c = to_a_phi (f->cx, k, f->n);  x += c.x * cos (a + c.y - phi);
        c = to_a_phi (f->cy, k, f->n);  y += c.x * cos (a + c.y);
    }
    
    return (p_t) { x, y };
}

void trafo_init_xy (trafo_t * f, spline_t * s)
{
    // f(l(t)) an M Stellen auswerten
    for (unsigned int k = 0; k < f->n; k++)
    {
        double t = (double) k / f->n;
        p_t p = spline_eval (s, t);
        f->x[k] = p.x;
        f->y[k] = p.y;
    }
}

void trafo_fft_xy (trafo_t * f)
{
    // FFT von fl bestimmen
    // FFT ist destruktiv auf der Eingabe
    gsl_fft_real_transform (f->x, 1, f->n, real, work);
    gsl_fft_real_transform (f->y, 1, f->n, real, work);

    // Die FFT kommt in einem gepackten Format an,
    // wir packen aus...
    gsl_fft_halfcomplex_unpack (f->x, f->cx, 1, f->n);
    gsl_fft_halfcomplex_unpack (f->y, f->cy, 1, f->n);
}

void trafo_fft (trafo_t * f, spline_t * s)
{
    trafo_init_xy (f, s);
    trafo_fft_xy (f);
}

// Kleine Koeffizienten vernachlssigen
int trafo_crop (trafo_t * f, double thres)
{
    int n = 0;
    
    for (size_t k=1; k <= f->n - k; k++)
    {
        if (to_a_phi (f->cx, k, f->n).x < thres) REAL(f->cx, k) = REAL (f->cx, f->n-k) = IMAG (f->cx, k) = IMAG (f->cx, f->n-k) = 0.0; else n++, f->m[0] = k+1;
        if (to_a_phi (f->cy, k, f->n).x < thres) REAL(f->cy, k) = REAL (f->cy, f->n-k) = IMAG (f->cy, k) = IMAG (f->cy, f->n-k) = 0.0; else n++, f->m[1] = k+1;
    }
    
    return n;
}

void trafo_dwt (trafo_t * f, spline_t * s)
{
    trafo_init_xy (f, s);
    gsl_wavelet_transform_forward (wave, f->x, 1, f->n, wwork);
    gsl_wavelet_transform_forward (wave, f->y, 1, f->n, wwork);
}

void trafo_dwt_xy (trafo_t * f)
{
    gsl_wavelet_transform_forward (wave, f->x, 1, f->n, wwork);
    gsl_wavelet_transform_forward (wave, f->y, 1, f->n, wwork);
}

void trafo_dwt_invert (trafo_t * f)
{
    gsl_wavelet_transform_inverse (wave, f->x, 1, f->n, wwork);
    gsl_wavelet_transform_inverse (wave, f->y, 1, f->n, wwork);
}

void trafo_plot_dwt (trafo_t * f, int color0, int color1, double max_t)
{
    max_t *= f->n;
    for (size_t k=0; k < f->n; k++)
        dot ((p_t) {f->x[k], f->y[k]}, k < max_t ? color0 : color1, 2);
}


void trafo_plot (trafo_t * f, int color0, int color1, double step, double max_t)
{
    max_t *= f->n;
    for (double t=0.0; t < f->n; t += step)
        dot (trafo_eval (f, t), t < max_t ? color0 : color1, 1);
}

// Zeichnet y[] als Funktion von t
void trafo_plot_y (trafo_t * f, int color)
{
    for (double x = -1.0; x <= 1; x += DPIX)
        dot ((p_t) {x, 0},   CL_BLACK, 1);
    
    for (size_t px = 0; px < f->n; px++)
        dot ((p_t) {-1.0+2.0*px/f->n, f->y[px] }, color, 1);
}

// Zeichnet cy[] als Funktion von t
void trafo_plot_cy (trafo_t * f, int color)
{
    for (double x = -1.0; x <= 1; x += DPIX)
        dot ((p_t) {x, 0},   CL_BLACK, 1);
    
    for (size_t px = 0; px < f->n; px++)
    {
        double t = -1.0+2.0*px/f->n;
        dot ((p_t) { t, trafo_eval (f, t).y }, color, 1);
    }
}

// Zeichnet x(t) und y(t) der FFT
void trafo_plot_xy (trafo_t * f, lambda_t * l, int colorx, int colory)
{
    // Senkrechte Linien
    for (size_t k = 0; k < l->n; k++)
    {
        double t = l->t[k];
        double x = -1+2*t;
        p_t p = trafo_eval (f, t*f->n);
        dot ((p_t) { x, p.x}, CL_RED, 4);
        dot ((p_t) { x, p.y}, CL_RED, 4);
        
        for (double y = -1.0; y <= 1; y += DPIX)
            dot ((p_t) { x, y }, CL_ROSE, 1);
    }

    for (double x = -1.0; x <= 1; x += DPIX)
    {
        double t = (x+1)/2 * f->n;
        p_t p = trafo_eval (f, t);
        dot ((p_t) {x, p.x}, colorx, 1);
        dot ((p_t) {x, p.y}, colory, 1);
        dot ((p_t) {x, 0},   CL_BLACK, 1);
    }
    
}


// Sucht besseres l, d.h. ein solches, bei dem
// die Bewertung der Trafo kleiner wird
int trafo_optimize (trafo_t * f, lambda_t * l, double *xdata, double *ydata, 
              void (*trans) (trafo_t*, spline_t*),
              double (*bewert) (trafo_t*))
{
    unsigned int j = 8;
    int change = 0;
    double delta = 0.05;
    lambda_t * l2 = lambda_alloc (l->n);
    spline_t *s   = spline_alloc (l->n, INTERPOL_SPLINE);
    double w, wert;
    
    spline_init (s, l, xdata, ydata);
    trans (f, s);
    wert = bewert (f);
    
    for (int i=0;; i++)
    {
        delta = -delta; // delta *= -0.997;
        int k = 1 + rand() % (l->n -1);
        if (i % 100 == 99) 
            j--, 
            delta *= 0.6;
        
        if (j == 0)
            break;
        double fak = 1.0;
retry:        
        if (!lambda_change (l2, l, k, j, fak*delta))
            continue;
        
        spline_init (s, l2, xdata, ydata);
        trans (f, s);
        w = bewert (f);
        
        if (w < wert)
        {
            change++;
            wert = w;
            lambda_copy (l, l2);
            fak /=2;
            if (fak > 1e-3)
                goto retry;
                
            i--;
            continue;
        }
    }
    
    spline_free (s);
    lambda_free (l2);
    
    return change;
}

static void trafo_out_kurve_C1 (FILE * file, double * c, size_t n, const char * s)
{
    size_t k, mk = 0;
    p_t c0 = to_a_phi (c, 0, n);
    
    fprintf (file, "\t%.3f,\t// %s0\n", c0.x * cos (c0.y), s);
    
    for (k=1; k <= n - k; k++)
        if (to_a_phi (c, k, n).x > 1e-3)
            mk = k;
            
    fprintf (file, "\t%d,\t// Anzahl %s-Terme\n", mk, s);
    for (k=1; k <= mk; k++)
    {
        p_t ck = to_a_phi (c, k, n);
        
        fprintf (file, "\t% .3f, %.3f,\t// A * cos (%d*t + phi), phi = % .4f * Pi\n", 
            ck.x, ck.y, k, ck.y/M_PI);
    }
    // LaTeX
    fprintf (file, "\n/*\n");
    fprintf (file, " %s0 = %.2f + \n", s, c0.x * cos (c0.y));
    for (k=1; k <= mk; k++)
    {
        p_t ck = to_a_phi (c, k, n);
        
        fprintf (file, " + \\ct{%.2f}{%d\\cdot t %.2f} \n", 
            ck.x, k, ck.y);
    }
    fprintf (file, "*/\n");
}

static void trafo_out_kurve_C_AVR1 (FILE * file, double * c, size_t n, const char * s)
{
    size_t k, mk = 0;
    p_t c0 = to_a_phi (c, 0, n);
    
    fprintf (file, "\t\tFRAC8 (%.3f),\t// %s0\n", c0.x * cos (c0.y), s);
    
    for (k=1; k <= n - k; k++)
        if (to_a_phi (c, k, n).x > 1e-3)
            mk = k;
            
    fprintf (file, "\t\t%d,\t// Anzahl %s-Terme\n", mk, s);
    for (k=1; k <= mk; k++)
    {
        p_t ck = to_a_phi (c, k, n);
        double phi = ck.y + M_PI/2;
        
        if (phi > M_PI*2)   phi -= M_PI*2;
        if (phi < 0)        phi += M_PI*2;
        
        unsigned short iphi = 360.*phi/M_PI;
        
        fprintf (file, "\t\tFRAC8 (%.3f), % 4d, %d%s\t// A * sin (%d*t + phi), phi = % .4f * 2*Pi\n", 
            ck.x, iphi & 0xff, iphi >> 8, k == mk && *s=='y' ? "" : ",", k, phi/(M_PI*2));
    }
}

// Trafo als C-Schnippsel  la kurve.h
void trafo_out_kurve_C (FILE * file, trafo_t * f, const char *name, double t_max)
{
    fprintf (file, "double %s[] =\n{\n", name);
    fprintf (file, "\t%.8f,\t// t in [0, %.8f * 2*Pi]\n", 2.0*M_PI*t_max, t_max);
    trafo_out_kurve_C1 (file, f->cx, f->n, "x");
    trafo_out_kurve_C1 (file, f->cy, f->n, "y");
    fprintf (file, "\t}\n};\n");
}

// Trafo als C-Schnippsel fr avr-gcc
void trafo_out_kurve_C_AVR (FILE * file, trafo_t * f, const char *name, double t_max)
{
    fprintf (file, "const uint8_t line_t %s PROGMEM =\n{\n", name);
    fprintf (file, "\t.t_max = %d,\t// t in [0, %.8f * 2*Pi]\n", (int) (720.*t_max), t_max);
    fprintf (file, "\t{\n");
    trafo_out_kurve_C_AVR1 (file, f->cx, f->n, "x");
    fprintf (file, "\t\n");
    trafo_out_kurve_C_AVR1 (file, f->cy, f->n, "y");
    fprintf (file, "\t}\n};\n");
}

static void trafo_dump1 (double * c, unsigned int n)
{
    for (unsigned int k=0; k <= n-k; k++)
    {
        p_t ck = to_a_phi (c, k, n);
        double a = ck.x / 1;
        if (a > 1e-5)
            printf (" % 2d: % .8f\n", k, a);
    }
}

// FFT austexten
void trafo_dump (const char *text, trafo_t * f)
{
    printf (text, "x"); trafo_dump1 (f->cx, f->n);
    printf (text, "y"); trafo_dump1 (f->cy, f->n);
    fflush (stdout);
}
